import asyncio
import os
import time
import threading
from signal import SIGINT, SIGTERM
from concurrent.futures import ThreadPoolExecutor
from aiorun import run


def kill(sig=SIGTERM, after=0.1):
    """Tool to send a signal after a pause"""
    def job():
        pid = os.getpid()
        time.sleep(after)
        os.kill(pid, sig)

    t = threading.Thread(target=job)
    t.start()


def test_sigterm():
    """Basic SIGTERM"""
    async def main():
        await asyncio.sleep(5.0)

    kill(SIGTERM)
    run(main())


def test_no_coroutine():
    """Signal should still work without a main coroutine"""
    kill(SIGTERM)
    run()


def test_sigint():
    """Basic SIGINT"""
    kill(SIGINT)
    run()


def test_exe():
    """Custom executor"""
    exe = ThreadPoolExecutor()
    kill(SIGTERM)
    run(executor=exe)


def test_sigint_pause():
    """Trying to send multiple signals."""
    async def main():
        try:
            await asyncio.sleep(5.0)
        except asyncio.CancelledError:
            await asyncio.sleep(0.2)

    kill(SIGINT, after=0.1)
    kill(SIGINT, after=0.15)
    run(main())



"""Boilerplate for asyncio applications"""
import logging
from asyncio import (
    get_event_loop,
    set_event_loop,
    new_event_loop,
    AbstractEventLoop,
    Task,
    gather,
    CancelledError,
)
from concurrent.futures import Executor, ThreadPoolExecutor
from signal import SIGTERM, SIGINT
from typing import Optional, Coroutine, Callable


__all__ = ['run']
__version__ = '2017.10.2'
logger = logging.getLogger('aiorun')


def shutdown():
    logger.debug('Entering shutdown handler')
    loop = get_event_loop()
    loop.remove_signal_handler(SIGTERM)
    loop.add_signal_handler(SIGINT, lambda: None)
    logger.critical('Stopping the loop')
    loop.stop()


def run(coro: Optional[Coroutine] = None, *,
        loop: Optional[AbstractEventLoop] = None,
        shutdown_handler: Callable[[], None] = shutdown,
        executor_workers: int = 10,
        executor: Optional[Executor] = None) -> None:
    logger.debug('Entering run()')
    if not loop:
        loop = new_event_loop()
        set_event_loop(loop)
    if coro:
        loop.create_task(coro)
    loop.add_signal_handler(SIGINT, shutdown_handler)
    loop.add_signal_handler(SIGTERM, shutdown_handler)
    if not executor:
        logger.debug('Creating default executor')
        executor = ThreadPoolExecutor(max_workers=executor_workers)
    loop.set_default_executor(executor)
    logger.critical('Running forever.')
    loop.run_forever()
    logger.critical('Entering shutdown phase.')
    tasks = Task.all_tasks()
    group = gather(*tasks)
    logger.critical('Cancelling pending tasks.')
    group.cancel()
    try:
        logger.critical('Running pending tasks till complete')
        loop.run_until_complete(group)
    except CancelledError:
        pass
    logger.critical('Waiting for executor shutdown.')
    executor.shutdown(wait=True)
    logger.critical('Closing the loop.')
    loop.close()
    logger.critical('Leaving. Bye!')
